#pragma config FEXTOSC = OFF    // FEXTOSC External Oscillator mode Selection bits (Oscillator not enabled)
#pragma config RSTOSC = HFINT1  // Power-up default value for COSC bits (HFINTOSC (1MHz))
#pragma config CLKOUTEN = OFF   // Clock Out Enable bit (CLKOUT function is disabled; I/O or oscillator function on OSC2)
#pragma config CSWEN = ON       // Clock Switch Enable bit (Writing to NOSC and NDIV is allowed)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)
#pragma config MCLRE = ON       // Master Clear Enable bit (MCLR/VPP pin function is MCLR; Weak pull-up enabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config WDTE = OFF       // Watchdog Timer Enable bits (WDT disabled; SWDTEN is ignored)
#pragma config LPBOREN = OFF    // Low-power BOR enable bit (ULPBOR disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable bits (Brown-out Reset disabled)
#pragma config PPS1WAY = ON     // PPSLOCK bit One-Way Set Enable bit (The PPSLOCK bit can be cleared and set only once; PPS registers remain locked after one clear/set cycle)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable bit (Stack Overflow or Underflow will cause a Reset)
#pragma config DEBUG = OFF      // Debugger enable bit (Background debugger disabled)
#pragma config WRT = OFF        // User NVM self-write protection bits (Write protection off)
#pragma config LVP = OFF        // Low Voltage Programming Enable bit (High Voltage on MCLR/VPP must be used for programming.)
#pragma config CP = OFF         // User NVM Program Memory Code Protection bit (User NVM code protection disabled)
#pragma config CPD = OFF        // Data NVM Memory Code Protection bit (Data NVM code protection disabled)

#include <xc.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


short timer;
#define _XTAL_FREQ  4000000     // 4MHz __delay_us()で使用
#define AGND        0x3C
#define temp_die    0x3D
#define DAC_O       0x3E
#define FVR         0x3F
#define ANA5        0x5
#define ANB0        0x8
#define ANC4        0x14
#define Manu        0x1
#define Auto        0x0
#define Power       0x1
#define Devi        0x0

unsigned char   stat = 0;   // SWの押下状態
char    time_up;            // タイマーのタイムアップ状態
short   ad_VR;              // VRのAD変換値
short   ad_TC;              // K熱電対のAD変換値
char    cnt1;
char    cnt2=0;
char    f_rcv;              // USART受信時のフラグ
char    rx_idx;             // 受信データのインデックス
char    rx_cnt;             // 受信データ数
char    rx_dat[32];         // 受信データ配列
float   Powm;               // 制御電力用
unsigned char Pow0;         // 制御電力用
char    mode;               // 動作モード
char    m_sel;              // メータ表示選択
float x,y;
float   I_ad;               // 積分値計算用
float   Ps = 0.008;         // PID動作時のP設定値例
//float   Ps = 10.0;          // ONOFF動作時のP設定値例
float   Is = 0.00001;         // PID動作時のI設定値例
//float   Is = 0.0;           // ONOFF動作時のI設定値例
float   Ds = 0.0;           // D設定用(未使用)
float   Ts = 200.0;         // 設定温度＝8.138mV(200℃)
short   ADs;                // 設定温度のAD換算値

short   TC[] = {   0,  397,  798, 1203, 1612,  2023,  2436,  2851,  3267,  3682,
                4096, 4509, 4920, 5328, 5735,  6138,  6540,  6941,  7340,  7739,
                8138, 8539, 8940, 9343, 9747, 10153, 10561, 10971, 11382, 11795};

short do_ADC(unsigned char adc_ch, unsigned char Vr){
    ADPCH = adc_ch;             // adc_chの設定
    ADREFbits.ADPREF = Vr;      // Vref+ の設定、Vr=0 VDD, Vr=3 FVR
    __delay_us( 5 );            // ADC条件が安定するまで5μs待つ
    ADCON0bits.ADGO = 1;        // AD変換開始
    while(ADCON0bits.ADGO == 1);// ADC終了まで待つ
    return ADRES;               // 変換結果をリターン
}

void ADC_Init(void)
{    
    ADCON0bits.ADCS = 0;        // ADC clk is Fosc
    ADCLKbits.ADCCS = 0x3f;     // ADC clk is Fosc/128=8MHz/128=62.5kHz(16uSec)
    ADREFbits.ADNREF = 0;       // Vref- is VSS
    ADCON0bits.ADFRM0 = 1;      // ADRESとADPREVのデータは右揃え
    ADCON0bits.ADCONT = 0;      // ADC終了時ADGOはクリア
    ADCON0bits.ADON = 1;        // ADC is enable  
}

void DAC1_Init(){
    DAC1CON0 = 0x8;             // DAC1PSS = 2, Vref = FVR
    DAC1CON0bits.DAC1EN = 1;    // DAC1動作
    DAC1CON0bits.DAC1OE1 = 1;   // DAC1出力はDAC1OUT1(RA2)端子
    TRISAbits.TRISA2 = 0;       // RA2端子は出力
}

void FVR_Init(){
    FVRCONbits.FVREN = 1;       // FVR動作
    while(!FVRCONbits.FVRRDY);  // FVRRDY = 1になるのを待つ
    FVRCONbits.ADFVR = 1;       // ADC Vref=1.024V
//    FVRCONbits.ADFVR = 2;       // ADC Vref=2.048V
    FVRCONbits.CDAFVR = 3;      // DAC Vref = 4.096V
} 

void EUSART_Init(){
    TX1STA = 0x20;
    RC1STA = 0x90;
        // BAUD = FOSC/16/SP1BRG
    BAUD1CON = 0x08;            // 16bit Baud
    SP1BRG = 12;                // BAUD RATE GENERATOR
    PIE3bits.RCIE = 1;
    PIR3bits.RCIF = 0;

    ANSELCbits.ANSC7 = 0;       // RC7 Digital
    TRISCbits.TRISC7 = 1;       // RC7 Input
    RC0PPS = 0x10;              // RC0PPS = TX/CK
}

void EUSART_Send(char c){
    while(TRMT == 0);
    TX1REG = c;
}
 
char EUSART_Receive(){
    if(PIR3bits.RCIF == 1){
        if((RC1STAbits.OERR)||(RC1STAbits.FERR)){
            RC1STA = 0;
            RC1STA = 0x90;
            return 0xff;
        }
        else
            return RC1REG;
    }
    else
        return 0;
}

void TMR1_Init()
{
    T1CON = 0b00110011;             //Prescale=1/8, RD16=1, ON=1
    T1CLK = 1;                      // Fosc/4
    TMR1 = 65536 - 12500;           // Cnt=12500
    PIR4bits.TMR1IF = 0;
    PIE4bits.TMR1IE = 1;
    INTCONbits.PEIE = 1;
}

void PWM7_Init()                    // PWM7 RC4出力
{                
    ANSELCbits.ANSC4 = 0;           // RC4 = Digital
    RC4PPS = 0x0F;                  // RC4 = PWM7OUT
    TRISCbits.TRISC4 = 0;           // RC4 = Output

    // TMR2 クロック=CLKR
    CLKRCON = 0b10010111;           // CLKREN = 1, CLKDRC = 10(50%), CLKRDIV = 7(1/128)
    CLKRCLK = 0;                    // Clock is FOSC = 4MHz
    PWM7DCH = 50;
    PIR4bits.TMR2IF = 0;
    T2CLKCON = 0b111;               // CLKR
    T2CONbits.CKPS = 0b111;         // 1:128
    PR2 = 249;                      // PWM7 = 4MHz / 128 / 128 / 250 = 0976Hz = 1.024s
                                    //        FOSC CLKRDIV CKPS PR2
    T2CONbits.ON = 1;               // Timer2 ON
    while(PIR4bits.TMR2IF == 0);    // Timer2起動?
    PWM7CONbits.PWM7EN = 1;         // PWM7イネーブル
}

short Calc_Ts(float T)
{
    short   a1;
    float   a2;
    float   Vs;
    
    a1 = (short)(T / 10.);                          // 一時使用 ex. Ts=206の時 a1=20
    a2 = fmod(Ts, 10.);                             // 一時使用 ex. Ts=206の時 a2=6
    Vs = TC[a1] + (TC[a1+1] - TC[a1]) * a2 / 10.;   // 設定温度のTC電圧計算
    return  Vs / 1000 * 101 / 1024 * 1023;          // 設定温度のAD値の計算
}

void interrupt isr()
{
    INTCONbits.GIE = 0;
    if(PIR4bits.TMR1IF == 1)        // TMR1割込み
    {
        cnt1++;
        rx_cnt++;
        TMR1 = 65536 - 12500;       // Cnt=12500
        if (cnt1 >= 5)              // 0.5s毎にタイムアップ
        {
            time_up = 1;
            cnt1 = 0;
            if (rx_cnt > 2)
            {
                rx_cnt = 2;
                rx_idx = 0;
            }
        }
        PIR4bits.TMR1IF = 0;
    }

    if(PIR3bits.RCIF == 1)          // USART受信割込み
    {
        if((RC1STAbits.OERR)||(RC1STAbits.FERR))
        {
            RC1STA = 0;
            RC1STA = 0x90;
        }
        else
        {
            rx_cnt = 0;
            rx_dat[rx_idx++] = RC1REG;
            if(rx_idx >= 16)
            {
                f_rcv = 1;
            }
        }
        PIR3bits.RCIF = 0;
    }
    INTCONbits.GIE = 1;
}
   

int main(void)
{
    OSCENbits.MFOEN = 1;
    OSCFRQ = 0x02;                  //4MHz
    OSCCON1bits.NDIV = 0x0;         //Clock Divider 1/1
    CLKRCON = 0x80+0x10;
    ANSELCbits.ANSC5 = 0;           // RC5 Digital
    ANSELCbits.ANSC6 = 0;           // RC6 Digital
    WPUC = 0x60;                    // RC5-6 = Weak Pull-up
    TRISC = 0xE0;                   // RC0-4 = Output,RC5-7 = Input
    LATC = 0x61;                    
    
    TMR1_Init();
    EUSART_Init();
    ADC_Init();
    DAC1_Init();
    FVR_Init();
    PWM7_Init();
    
    INTCONbits.GIE = 1;
    
    ADs = Calc_Ts(Ts);              // 設定温度のAD値を求める
    
    while (1) 
    {
        if (time_up)
        {
            time_up = 0;
            cnt2++;
            if (RC5 == 1)           // Auto/Manu? 
            {
                mode = Auto;
                stat |= 1;
            }
            else 
            {
                mode = Manu;
                stat &= ~1;
            } 
            if (RC6 == 1)           // 偏差/出力表示？
            {
                m_sel = Devi;
                stat |= 2;
            } 
            else 
            {
                m_sel = Power;
                stat &= ~2;
            } 

            if (mode == Manu)
            {
                ad_VR = do_ADC(ANB0, 0);  // VR VDD
                Pow0 = (float)ad_VR / 1023. * 249.;
                PWM7DCH = (char)Pow0;
                x=(float)Pow0/2.50;
                y = 3E-05*pow(x,3) - 0.0023*pow(x,2) + 0.2073*x - 0.1242;
                DAC1CON1 = (char)y;        // DAC1R ????
            }
            else
            {
                ad_TC = do_ADC(ANA5, 3);        // TCのAD変換 FVR_buffer1
                if (abs(ADs - ad_TC) > 100)     // 設定値と実際の温度が100以上違う場合は積分値をリセット
                {
                    I_ad = 0;
                }
                I_ad = I_ad + ADs - ad_TC;      // 積分値を求める
                Powm = ((ADs - ad_TC) * Ps + I_ad * Is) * 249.; // 出力を求める
                Pow0 = Powm;
                if (Powm < 0)                   // 0<=出力<=249
                {
                    Pow0 = 0;
                }
                if (Powm > 249)
                {
                    Pow0 = 249;
                }
                PWM7DCH = Pow0;                 // 出力をPWMに設定
                
            }
            ad_TC = do_ADC(ANA5, 3);        // TCのAD変換 FVR_buffer1

            if (cnt2 >= 2)              // 0.5s×2＝1s毎にデータをPCへ送信
            {
                cnt2 = 0;
                EUSART_Send(stat);
                EUSART_Send(ad_TC >> 8);
                EUSART_Send(ad_TC);
                EUSART_Send(0);
                EUSART_Send(Pow0);
            }
            
            if (f_rcv == 1)                     // PCからTsなどの設定値
            {
                f_rcv = 0;
                rx_idx = 0;
                Ts = (rx_dat[0] * 256. + rx_dat[1]) / 10.;  // 設定温度の取得
                ADs = Calc_Ts(Ts);              // 設定温度のAD値を求める
                Ps = (rx_dat[2] * 16777216. + rx_dat[3] * 65536. + rx_dat[4] * 256. + rx_dat[5]) / 10000000.;
                Is = (rx_dat[6] * 16777216. + rx_dat[7] * 65536. + rx_dat[8] * 256. + rx_dat[9]) / 10000000.;
                Ds = (rx_dat[10] * 16777216. + rx_dat[11] * 65536. + rx_dat[12] * 256. + rx_dat[13]) / 10000000.;
                
            }
        }

    }
}

